﻿## **Vario** • Transactional, Reactive, and Asynchronous State Management for JavaScript

Inspired by: MobX, Nezaboodka, React, Excel.

## Introduction

Vario is a transactional, reactive, and asynchronous state
management library for JavaScript that is designed to be
extermely lightweight, easy, and fast.

**Transactivity** means that multiple objects can be changed
at once with full respect to the all-or-nothing principle
(atomicity, consistency, and isolation). Vario maintains
separate data snapshot for each transaction. The snapshot
is logical and doesn't create full copy of all the data.
Intermediate state is visible only inside transaction itself,
but is not visible outside of transaction until it is committed.
Compensating actions are not needed in case of transaction
failure, because all the changes made by transaction in its
logical snapshot are simply discarded.

**Reactivity** means that recomputation of computable objects
(observers) is triggered automatically upon changes in
their dependencies (observables). All the dependencies
between observers and their observables are detected and
maintained automatically. It is achieved by injecting
property getters/setters into all objects and tracking
get/set calls during execution of observer computation.
Affected observers are recomputed in a proper order at the
end of a transaction, when all the changes are committed.

**Asynchrony** means that asynchronous operations are supported
as first class citizens during transaction processing.
Transaction may consist of a set of asynchronous operations
and being committed upon completion of all them. Moreover,
any asynchronous operation may spawn other asynchronous
operations, which prolong transaction execution until whole
the chain of asynchronous operations is fully completed.

## Differentiators

* Consistency and clarity are the first priorities
* __Transactional__ -- full-fledged atomicity, consistency, and isolation
* __Reactive__ -- automatic dependency tracking and fine-grained recomputation
* __Asynchronous__ -- transaction may consist of parallel and chained asynchronous operations
* __Historical__ -- built-in undo/redo functionality provided out of the box
* __Minimalistic__ -- it's a tool and approach, not a framework
* __Trivial__ -- implementation consists of less than 1000 lines of code

## Demo

```typescript
import { Transaction, tran, cache, autorenew } from "vario";
import { Person } from "./person";

@tran
export class DemoApp {
  @tran title: string = "Demo";
  @tran users: Person[] = [];

  @tran
  loadUsers(): void {
    this.users.push(new Person({
      name: "John", age: 38,
      emails: ["john@mail.com"],
      children: [
        new Person({ name: "Billy" }), // William
        new Person({ name: "Barry" }), // Barry
        new Person({ name: "Steve" }), // Steven
      ],
    }));
    this.users.push(new Person({
      name: "Kevin", age: 27,
      emails: ["kevin@mail.com"],
      children: [
        new Person({ name: "Britney" }),
      ],
    }));
  }
}

@tran
export class DemoAppView {
  readonly model: DemoApp;
  @tran userFilter: string = "Jo";

  constructor(model: DemoApp) {
    this.model = model;
  }

  @cache
  filteredUsers(): Person[] {
    const m = this.model;
    let result: Person[] = m.users;
    if (this.userFilter.length > 0) {
      result = [];
      for (let x of m.users)
        if (x.name && x.name.indexOf(this.userFilter) === 0)
          result.push(x);
    }
    return result;
  }

  @cache
  render(): string[] {
    // Print only those users whos name starts with filter string
    let r: string[] = [];
    r.push("---");
    r.push(`Filter: ${this.userFilter}`);
    const a = this.filteredUsers();
    for (let x of a) {
      let childNames = x.children.map(child => child.name);
      r.push(`${x.name}'s children: ${childNames.join(", ")}`);
    }
    r.push("---");
    return r;
  }

  @cache
  print(): void {
    this.render().forEach(x => console.log(x));
  }
}

export function sample(): void {
  // Simple actions (transactions)
  let app = new DemoApp();
  let view = new DemoAppView(app);
  try {
    app.loadUsers();
    autorenew(0, view.print);
    // Multi-part transaction
    let t1 = new Transaction("t1");
    t1.run(() => {
      let daddy = app.users[0];
      daddy.age += 2; // causes no execution of DemoApp.render
      daddy.name = "John Smith"; // causes execution of DemoApp.render upon transaction end
      daddy.children[0].name = "Barry Smith";   // Barry
      daddy.children[1].name = "William Smith"; // Billy
      daddy.children[2].name = "Steven Smith";  // Steve
    });
    t1.run(() => {
      // daddy.age is 38 outside of t2 transaction, but is 40 inside t2
      let daddy = app.users[0];
      daddy.age += 5; // 40 + 5 = 45
      view.userFilter = "";
      if (daddy.emails)
        daddy.emails[0] = "daddy@mail.com";
      let x = daddy.children[1];
      x.parent = null;
      x.parent = daddy;
    });
    t1.commit(); // changes are applied, caches are invalidated/renewed
    // Protection from modification outside of a transaction
    try {
      let daddy = app.users[0];
      if (daddy.emails)
        daddy.emails.push("dad@mail.com");
      else
        daddy.children[1].name = "Billy Smithy";
    }
    catch (e) {
      console.log(`Expected: ${e}`);
    }
    // Turn off auto renew
    autorenew(-1, view.print);
  }
  finally { // cleanup
    // dispose(view);
    // dispose(app);
  }
}

/* Console output:

#vario t11 ╔═══ v10 DemoApp.ctor - begin
#vario t11 ║ M DemoApp#11t11: title, users
#vario t11 ╚═══ v11 DemoApp.ctor - end (COMMIT)
#vario t12 ╔═══ v11 DemoAppView.ctor - begin
#vario t12 ║ M DemoAppView#12t12: userFilter
#vario t12 ╚═══ v12 DemoAppView.ctor - end (COMMIT)
#vario t13 ╔═══ v12 DemoApp#11t10.loadUsers - begin
#vario t13 ║ ┌── DemoApp#11t13.loadUsers - begin
#vario t13 ║ │ ┌── Person#16t13.appendChildren - begin
#vario t13 ║ │ │ ┌── Person#13t13.setParent - begin
#vario t13 ║ │ │ └── Person#13t13.setParent - end
#vario t13 ║ │ │ ┌── Person#14t13.setParent - begin
#vario t13 ║ │ │ └── Person#14t13.setParent - end
#vario t13 ║ │ │ ┌── Person#15t13.setParent - begin
#vario t13 ║ │ │ └── Person#15t13.setParent - end
#vario t13 ║ │ └── Person#16t13.appendChildren - end
#vario t13 ║ │ ┌── Person#18t13.appendChildren - begin
#vario t13 ║ │ │ ┌── Person#17t13.setParent - begin
#vario t13 ║ │ │ └── Person#17t13.setParent - end
#vario t13 ║ │ └── Person#18t13.appendChildren - end
#vario t13 ║ └── DemoApp#11t13.loadUsers - end
#vario t13 ║ M DemoApp#11t13: users
#vario t13 ║ M Person#13t13: id, name, age, emails, log, _parent, _children
#vario t13 ║ M Person#14t13: id, name, age, emails, log, _parent, _children
#vario t13 ║ M Person#15t13: id, name, age, emails, log, _parent, _children
#vario t13 ║ M Person#16t13: id, name, age, emails, log, _parent, _children
#vario t13 ║ M Person#17t13: id, name, age, emails, log, _parent, _children
#vario t13 ║ M Person#18t13: id, name, age, emails, log, _parent, _children
#vario t13 ╚═══ v13 DemoApp#11t10.loadUsers - end (COMMIT)
#vario t14 ╔═══ v13 DemoAppView#12t10.print - begin
#vario t14 ║ ┌── DemoAppView#12t14.print - begin
#vario t14 ║ │ ┌── DemoAppView#12t14.render - begin
#vario t14 ║ │ │ ∞ DemoAppView#12t14.render uses DemoAppView#12t14.userFilter
#vario t14 ║ │ │ ┌── DemoAppView#12t14.filteredUsers - begin
#vario t14 ║ │ │ │ ∞ DemoAppView#12t14.filteredUsers uses DemoApp#11t13.users
#vario t14 ║ │ │ │ ∞ DemoAppView#12t14.filteredUsers uses DemoAppView#12t14.userFilter
#vario t14 ║ │ │ │ ∞ DemoAppView#12t14.filteredUsers uses DemoApp#11t13.users
#vario t14 ║ │ │ │ ∞ DemoAppView#12t14.filteredUsers uses Person#16t13.name
#vario t14 ║ │ │ │ ∞ DemoAppView#12t14.filteredUsers uses Person#16t13.name
#vario t14 ║ │ │ │ ∞ DemoAppView#12t14.filteredUsers uses DemoAppView#12t14.userFilter
#vario t14 ║ │ │ │ ∞ DemoAppView#12t14.filteredUsers uses Person#18t13.name
#vario t14 ║ │ │ │ ∞ DemoAppView#12t14.filteredUsers uses Person#18t13.name
#vario t14 ║ │ │ │ ∞ DemoAppView#12t14.filteredUsers uses DemoAppView#12t14.userFilter
#vario t14 ║ │ │ └── DemoAppView#12t14.filteredUsers - end
#vario t14 ║ │ │ ∞ DemoAppView#12t14.render uses DemoAppView#12t14.filteredUsers
#vario t14 ║ │ │ ∞ DemoAppView#12t14.render uses Person#16t13._children
#vario t14 ║ │ │ ∞ DemoAppView#12t14.render uses Person#13t13.name
#vario t14 ║ │ │ ∞ DemoAppView#12t14.render uses Person#14t13.name
#vario t14 ║ │ │ ∞ DemoAppView#12t14.render uses Person#15t13.name
#vario t14 ║ │ │ ∞ DemoAppView#12t14.render uses Person#16t13.name
#vario t14 ║ │ └── DemoAppView#12t14.render - end
#vario t14 ║ │ ∞ DemoAppView#12t14.print uses DemoAppView#12t14.render
---
Filter: Jo
John's children: Billy, Barry, Steve
---
#vario t14 ║ └── DemoAppView#12t14.print - end
#vario t14 ║ M DemoAppView#12t14: filteredUsers, render
#vario t14 ╚═══ v14 DemoAppView#12t10.print - end (COMMIT)
#vario t15 ╔═══ v14 t1 - begin
#vario t15 ║ ┌── Person#14t15.setParent - begin
#vario t15 ║ └── Person#14t15.setParent - end
#vario t15 ║ ┌── Person#14t15.setParent - begin
#vario t15 ║ └── Person#14t15.setParent - end
#vario t15 ║ M Person#16t15: age, name, emails, _children
#vario t15 ║ M Person#13t15: name
#vario t15 ║ M Person#14t15: name
#vario t15 ║ M Person#15t15: name
#vario t15 ║ M DemoAppView#12t15: userFilter
#vario t15 ╚═══ v15 t1 - end (COMMIT)
#vario t15   i DemoAppView#12t14.filteredUsers is invalidated by Person#16t15.name
#vario t15   i DemoAppView#12t14.render is invalidated by DemoAppView#12t14.filteredUsers
#vario t15   ■ DemoAppView#12t14.print is invalidated by DemoAppView#12t14.render
#vario t16 ╔═══ v15 DemoAppView#12t14.print - begin
#vario t16 ║ ┌── DemoAppView#12t16.print - begin
#vario t16 ║ │ ┌── DemoAppView#12t16.render - begin
#vario t16 ║ │ │ ∞ DemoAppView#12t16.render uses DemoAppView#12t16.userFilter
#vario t16 ║ │ │ ┌── DemoAppView#12t16.filteredUsers - begin
#vario t16 ║ │ │ │ ∞ DemoAppView#12t16.filteredUsers uses DemoApp#11t13.users
#vario t16 ║ │ │ │ ∞ DemoAppView#12t16.filteredUsers uses DemoAppView#12t16.userFilter
#vario t16 ║ │ │ └── DemoAppView#12t16.filteredUsers - end
#vario t16 ║ │ │ ∞ DemoAppView#12t16.render uses DemoAppView#12t16.filteredUsers
#vario t16 ║ │ │ ∞ DemoAppView#12t16.render uses Person#16t15._children
#vario t16 ║ │ │ ∞ DemoAppView#12t16.render uses Person#13t15.name
#vario t16 ║ │ │ ∞ DemoAppView#12t16.render uses Person#15t15.name
#vario t16 ║ │ │ ∞ DemoAppView#12t16.render uses Person#14t15.name
#vario t16 ║ │ │ ∞ DemoAppView#12t16.render uses Person#16t15.name
#vario t16 ║ │ │ ∞ DemoAppView#12t16.render uses Person#18t13._children
#vario t16 ║ │ │ ∞ DemoAppView#12t16.render uses Person#17t13.name
#vario t16 ║ │ │ ∞ DemoAppView#12t16.render uses Person#18t13.name
#vario t16 ║ │ └── DemoAppView#12t16.render - end
#vario t16 ║ │ ∞ DemoAppView#12t16.print uses DemoAppView#12t16.render
---
Filter:
John Smith's children: Barry Smith, Steven Smith, William Smith
Kevin's children: Britney
---
#vario t16 ║ └── DemoAppView#12t16.print - end
#vario t16 ║ M DemoAppView#12t16: filteredUsers, render
#vario t16 ╚═══ v16 DemoAppView#12t14.print - end (COMMIT)
Expected: Error: E609: object cannot be changed outside of transaction

*/
```

## Async Demo

```typescript
import { tran, cache, autorenew } from "vario";
import { setTimeout } from "timers";
import fetch from "node-fetch";

@tran
export class DemoApp {
  @tran title: string = "Demo";
  @tran items: string[] = [];

  @tran
  async download(url: string, delay: number): Promise<void> {
    this.title = "Demo (" + new Date().toISOString() + ")";
    let start = Date.now();
    await all([fetch(url), sleep(delay)]);
    let ms = Date.now() - start;
    this.items.push(`${url} in ${ms} ms`);
  }
}

export class DemoAppView {
  readonly model: DemoApp;

  constructor(model: DemoApp) {
    this.model = model;
  }

  @cache
  async render(): Promise<string[]> {
    let r: string[] = [];
    r.push("---");
    r.push("Title: " + this.model.title);
    await sleep(1000);
    r.push("Items: ");
    for (let x of this.model.items)
      r.push(" - " + x);
    r.push("---");
    return r;
  }

  @cache
  async print(): Promise<void> {
    let lines: string[] = await this.render();
    lines.forEach(x => console.log(x));
  }
}

export async function sample(): Promise<void> {
  let app = new DemoApp();
  let view = new DemoAppView(app);
  try {
    autorenew(0, view.print);
    let list: Array<{ url: string, delay: number }> = [
      { url: "https://nezaboodka.com", delay: 700 },
      { url: "https://google.com", delay: 2000 },
      { url: "https://microsoft.com", delay: 700 },
    ];
    await all(list.map(x => app.download(x.url, x.delay)));
    autorenew(-1, view.print);
  }
  catch (error) {
    console.log(`${error}`);
  }
  finally {
    // dispose(view);
    // dispose(app);
  }
}

async function sleep(timeout: number): Promise<void> {
  return new Promise<void>(function(resolve) {
    setTimeout(resolve.bind(null, () => resolve), timeout);
  });
}

async function all(promises: Array<Promise<any>>): Promise<any[]> {
  let error: any;
  let result = await Promise.all(promises.map(x => x.catch(e => { error = error || e; return e; })));
  if (error)
    throw error;
  return result;
}

/* Console output:

#vario t17 ╔═══ v16 DemoApp.ctor - begin
#vario t17 ║ M DemoApp#19t17: title, items
#vario t17 ╚═══ v17 DemoApp.ctor - end (COMMIT)
#vario t18 ╔═══ v17 DemoAppView#20t10.print - begin
#vario t18 ║ ┌── DemoAppView#20t18.print - begin
#vario t18 ║ │ ┌── DemoAppView#20t18.render - begin
#vario t18 ║ │ │ ∞ DemoAppView#20t18.render uses DemoApp#19t17.title
#vario t19 ╔═══ v17 DemoApp#19t10.download - begin
#vario t19 ║ ┌── DemoApp#19t19.download - begin
#vario t20 ╔═══ v17 DemoApp#19t10.download - begin
#vario t20 ║ ┌── DemoApp#19t20.download - begin
#vario t21 ╔═══ v17 DemoApp#19t10.download - begin
#vario t21 ║ ┌── DemoApp#19t21.download - begin
#vario t19 ║ └── DemoApp#19t19.download - end
#vario t19 ║ M DemoApp#19t19: title, items
#vario t19 ╚═══ v18 DemoApp#19t10.download - end (COMMIT)
#vario t21 ║ └── DemoApp#19t21.download - end
#vario t21 ║ M DemoApp#19t21: title, items
#vario t21 ╚═══ v17 DemoApp#19t10.download - end (DISCARD) Error: DemoApp#19t10.download conflicts with other transactions on: DemoApp#19t19.title, DemoApp#19t19.items
#vario t18 ║ │ │ ∞ DemoAppView#20t18.render uses DemoApp#19t17.items (DIRTY)
#vario t18 ║ │ └── DemoAppView#20t18.render - end
#vario t18 ║ │ ∞ DemoAppView#20t18.print uses DemoAppView#20t18.render
---
Title: Demo
Items:
---
#vario t18 ║ └── DemoAppView#20t18.print - end
#vario t18 ║ M DemoAppView#20t18: render
#vario t18 ╚═══ v19 DemoAppView#20t10.print - end (COMMIT)
#vario t18   i DemoAppView#20t18.render is invalidated by DemoApp#19t17.items
#vario t18   ■ DemoAppView#20t18.print is invalidated by DemoAppView#20t18.render
#vario t22 ╔═══ v19 DemoAppView#20t18.print - begin
#vario t22 ║ ┌── DemoAppView#20t22.print - begin
#vario t22 ║ │ ┌── DemoAppView#20t22.render - begin
#vario t22 ║ │ │ ∞ DemoAppView#20t22.render uses DemoApp#19t19.title
#vario t20 ║ └── DemoApp#19t20.download - end
#vario t20 ║ M DemoApp#19t20: title, items
#vario t20 ╚═══ v17 DemoApp#19t10.download - end (DISCARD) Error: DemoApp#19t10.download conflicts with other transactions on: DemoApp#19t19.title, DemoApp#19t19.items
Error: DemoApp#19t10.download conflicts with other transactions on: DemoApp#19t19.title, DemoApp#19t19.items
#vario t22 ║ │ │ ∞ DemoAppView#20t22.render uses DemoApp#19t19.items
#vario t22 ║ │ └── DemoAppView#20t22.render - end
#vario t22 ║ │ ∞ DemoAppView#20t22.print uses DemoAppView#20t22.render
---
Title: Demo (2018-10-02T15:40:30.915Z)
Items:
 - https://nezaboodka.com in 722 ms
---
#vario t22 ║ └── DemoAppView#20t22.print - end
#vario t22 ║ M DemoAppView#20t22: render
#vario t22 ╚═══ v20 DemoAppView#20t18.print - end (COMMIT)

*/
```

## API (TypeScript)

```typescript

// Decorators

export type F<T> = (...args: any[]) => T;
export const transactional = tran;
export function tran(target: object, prop?: string, descriptor?: TypedPropertyDescriptor<F<any>>): any;
export function cache(target: Object, prop: string, descriptor: TypedPropertyDescriptor<F<any>>): any;

// Control functions

export interface Status { value: any; outdated: boolean; error: any; latency: number; }
export function statusof(method: F<any>): Status;
export function autorenew(latency: number, method: F<any>): void;
export function dispose(obj: object | undefined): void;
  
// Transaction

export class Transaction {
  constructor(hint?: string);
  run<T>(func: F<T>, ...args: any[]): T;
  wrap<T>(func: F<T>): F<T>;
  commit(): void;
  sealToCommit(): Transaction; // t1.sealToCommit().waitForFinish().then(fulfill, reject)
  discard(error?: any): Transaction; // t1.sealToCommit().waitForFinish().then(...)
  waitForFinish(): Promise<void>;
  finished(): boolean;
  undo(): void; // not implemented
  redo(): void; // not implemented
  static run<T>(hint: string, func: F<T>, ...args: any[]): T;
  static async runAsync<T>(hint: string, func: F<Promise<T>>, ...args: any[]): Promise<T>;
  static get current(): Transaction;
}

```
